@@ -1,22 +1,22 @@
-import threading
-from util.threadinglocal import local as threading_local
-import sys
-import re
-import warnings
 import atexit
-import os
+from cgi import parse_qsl
+import inspect
 import new
+import os
+import sys
+import threading
 import types
 import urllib
+import warnings
 import weakref
-import inspect
-import sqlbuilder
+
 from cache import CacheSet
-import col
-import main
-from joins import sorter
-from converters import sqlrepr
 import classregistry
+import col
+from converters import sqlrepr
+import main
+import sqlbuilder
+from util.threadinglocal import local as threading_local
 
 warnings.filterwarnings("ignore", "DB-API extension cursor.lastrowid used")
 
@@ -28,26 +28,33 @@ def _closeConnection(ref):
         conn.close()
 
 class ConsoleWriter:
-    def __init__(self, loglevel):
-        self.loglevel = loglevel
-        self.logfile = getattr(sys, loglevel or "stdout")
+    def __init__(self, connection, loglevel):
+        # loglevel: None or empty string for stdout; or 'stderr'
+        self.loglevel = loglevel or "stdout"
+        self.dbEncoding = getattr(connection, "dbEncoding", None) or "ascii"
     def write(self, text):
-        self.logfile.write(text + '\n')
+        logfile = getattr(sys, self.loglevel)
+        if isinstance(text, unicode):
+            try:
+                text = text.encode(self.dbEncoding)
+            except UnicodeEncodeError:
+                text = repr(text)[2:-1] # Remove u'...' from the repr
+        logfile.write(text + '\n')
 
 class LogWriter:
-    def __init__(self, logger, loglevel):
+    def __init__(self, connection, logger, loglevel):
         self.logger = logger
         self.loglevel = loglevel
         self.logmethod = getattr(logger, loglevel)
     def write(self, text):
         self.logmethod(text)
 
-def makeDebugWriter(loggerName, loglevel):
+def makeDebugWriter(connection, loggerName, loglevel):
     if not loggerName:
-        return ConsoleWriter(loglevel)
+        return ConsoleWriter(connection, loglevel)
     import logging
     logger = logging.getLogger(loggerName)
-    return LogWriter(logger, loglevel)
+    return LogWriter(connection, logger, loglevel)
 
 class Boolean(object):
     """A bool class that also understands some special string keywords (yes/no, true/false, on/off, 1/0)"""
@@ -69,7 +76,7 @@ class DBConnection:
         self.debug = Boolean(debug)
         self.debugOutput = Boolean(debugOutput)
         self.debugThreading = Boolean(debugThreading)
-        self.debugWriter = makeDebugWriter(logger, loglevel)
+        self.debugWriter = makeDebugWriter(self, logger, loglevel)
         self.doCache = Boolean(cache)
         self.cache = CacheSet(cache=self.doCache)
         self.style = style
@@ -77,13 +84,12 @@ class DBConnection:
         self._connectionCount = 1
         self.autoCommit = Boolean(autoCommit)
         self.registry = registry or None
-        classregistry.registry(self.registry).addCallback(
-            self.soClassAdded)
+        classregistry.registry(self.registry).addCallback(self.soClassAdded)
         registerConnectionInstance(self)
         atexit.register(_closeConnection, weakref.ref(self))
 
-    def uri(self):
-        auth = getattr(self, 'user', '')
+    def oldUri(self):
+        auth = getattr(self, 'user', '') or ''
         if auth:
             if self.password:
                 auth = auth + ':' + self.password
@@ -94,21 +100,45 @@ class DBConnection:
         uri = '%s://%s' % (self.dbName, auth)
         if self.host:
             uri += self.host
+            if self.port:
+                uri += ':%d' % self.port
         uri += '/'
         db = self.db
         if db.startswith('/'):
-            db = path[1:]
+            db = db[1:]
         return uri + db
 
-    def isSupported(cls):
-        raise NotImplemented
-    isSupported = classmethod(isSupported)
+    def uri(self):
+        auth = getattr(self, 'user', '') or ''
+        if auth:
+            auth = urllib.quote(auth)
+            if self.password:
+                auth = auth + ':' + urllib.quote(self.password)
+            auth = auth + '@'
+        else:
+            assert not getattr(self, 'password', None), (
+                'URIs cannot express passwords without usernames')
+        uri = '%s://%s' % (self.dbName, auth)
+        if self.host:
+            uri += self.host
+            if self.port:
+                uri += ':%d' % self.port
+        uri += '/'
+        db = self.db
+        if db.startswith('/'):
+            db = db[1:]
+        return uri + urllib.quote(db)
 
+    @classmethod
+    def connectionFromOldURI(cls, uri):
+        return cls._connectionFromParams(*cls._parseOldURI(uri))
+
+    @classmethod
     def connectionFromURI(cls, uri):
-        raise NotImplemented
-    connectionFromURI = classmethod(connectionFromURI)
+        return cls._connectionFromParams(*cls._parseURI(uri))
 
-    def _parseURI(uri):
+    @staticmethod
+    def _parseOldURI(uri):
         schema, rest = uri.split(':', 1)
         assert rest.startswith('/'), "URIs must start with scheme:/ -- you did not include a / (in %r)" % rest
         if rest.startswith('/') and not rest.startswith('//'):
@@ -125,8 +155,7 @@ class DBConnection:
             else:
                 host, rest = rest.split('/', 1)
         if host and host.find('@') != -1:
-            user = host[:host.rfind('@')] # Python 2.3 doesn't have .rsplit()
-            host = host[host.rfind('@')+1:] # !!!
+            user, host = host.rsplit('@', 1)
             if user.find(':') != -1:
                 user, password = user.split(':', 1)
             else:
@@ -157,7 +186,45 @@ class DBConnection:
                 argvalue = urllib.unquote(argvalue)
                 args[argname] = argvalue
         return user, password, host, port, path, args
-    _parseURI = staticmethod(_parseURI)
+
+    @staticmethod
+    def _parseURI(uri):
+        protocol, request = urllib.splittype(uri)
+        user, password, port = None, None, None
+        host, path = urllib.splithost(request)
+
+        if host:
+            # Python < 2.7 have a problem - splituser() calls unquote() too early
+            #user, host = urllib.splituser(host)
+            if '@' in host:
+                user, host = host.split('@', 1)
+            if user:
+                user, password = [x and urllib.unquote(x) or None for x in urllib.splitpasswd(user)]
+            host, port = urllib.splitport(host)
+            if port: port = int(port)
+        elif host == '':
+            host = None
+
+        # hash-tag is splitted but ignored
+        path, tag = urllib.splittag(path)
+        path, query = urllib.splitquery(path)
+
+        path = urllib.unquote(path)
+        if (os.name == 'nt') and (len(path) > 2):
+            # Preserve backward compatibility with URIs like /C|/path;
+            # replace '|' by ':'
+            if path[2] == '|':
+                path = "%s:%s" % (path[0:2], path[3:])
+            # Remove leading slash
+            if (path[0] == '/') and (path[2] == ':'):
+                path = path[1:]
+
+        args = {}
+        if query:
+            for name, value in parse_qsl(query):
+                args[name] = value
+
+        return user, password, host, port, path, args
 
     def soClassAdded(self, soClass):
         """
@@ -578,7 +645,10 @@ class DBAPI(DBConnection):
             data[soClass.sqlmeta.idName] = kw.pop('id')
         for key, col in soClass.sqlmeta.columns.items():
             if key in kw:
-                data[col.dbName] = kw.pop(key)
+                value = kw.pop(key)
+                if col.from_python:
+                    value = col.from_python(value, sqlbuilder.SQLObjectState(soClass, connection=self))
+                data[col.dbName] = value
             elif col.foreignName in kw:
                 obj = kw.pop(col.foreignName)
                 if isinstance(obj, main.SQLObject):
@@ -612,6 +682,8 @@ class DBAPI(DBConnection):
             return
         self._poolLock.acquire()
         try:
+            if not self._pool: # _pool could be filled in a different thread
+                return
             conns = self._pool[:]
             self._pool[:] = []
             for conn in conns:
@@ -716,7 +788,7 @@ class Transaction(object):
 
     def _SO_delete(self, inst):
         cls = inst.__class__.__name__
-        if not self._deletedCache.has_key(cls):
+        if not cls in self._deletedCache:
             self._deletedCache[cls] = []
         self._deletedCache[cls].append(inst.id)
         meth = new.instancemethod(self._dbConnection._SO_delete.im_func, self, self.__class__)
@@ -827,7 +899,7 @@ class ConnectionHub(object):
         # I'm a little surprised we have to do this, but apparently
         # the object's private dictionary of attributes doesn't
         # override this descriptor.
-        if obj and obj.__dict__.has_key('_connection'):
+        if (obj is not None) and '_connection' in obj.__dict__:
             return obj.__dict__['_connection']
         return self.getConnection()
 
@@ -903,41 +975,42 @@ class ConnectionURIOpener(object):
 
     def __init__(self):
         self.schemeBuilders = {}
-        self.schemeSupported = {}
         self.instanceNames = {}
         self.cachedURIs = {}
 
-    def registerConnection(self, schemes, builder, isSupported):
+    def registerConnection(self, schemes, builder):
         for uriScheme in schemes:
-            assert not self.schemeBuilders.has_key(uriScheme) \
+            assert not uriScheme in self.schemeBuilders \
                    or self.schemeBuilders[uriScheme] is builder, \
                    "A driver has already been registered for the URI scheme %s" % uriScheme
             self.schemeBuilders[uriScheme] = builder
-            self.schemeSupported = isSupported
 
     def registerConnectionInstance(self, inst):
         if inst.name:
-            assert not self.instanceNames.has_key(inst.name) \
+            assert not inst.name in self.instanceNames \
                    or self.instanceNames[inst.name] is cls, \
                    "A instance has already been registered with the name %s" % inst.name
             assert inst.name.find(':') == -1, "You cannot include ':' in your class names (%r)" % cls.name
             self.instanceNames[inst.name] = inst
 
-    def connectionForURI(self, uri, **args):
+    def connectionForURI(self, uri, oldUri=False, **args):
         if args:
             if '?' not in uri:
                 uri += '?' + urllib.urlencode(args)
             else:
                 uri += '&' + urllib.urlencode(args)
-        if self.cachedURIs.has_key(uri):
+        if uri in self.cachedURIs:
             return self.cachedURIs[uri]
         if uri.find(':') != -1:
             scheme, rest = uri.split(':', 1)
             connCls = self.dbConnectionForScheme(scheme)
-            conn = connCls.connectionFromURI(uri)
+            if oldUri:
+                conn = connCls.connectionFromOldURI(uri)
+            else:
+                conn = connCls.connectionFromURI(uri)
         else:
             # We just have a name, not a URI
-            assert self.instanceNames.has_key(uri), \
+            assert uri in self.instanceNames, \
                    "No SQLObject driver exists under the name %s" % uri
             conn = self.instanceNames[uri]
         # @@: Do we care if we clobber another connection?
@@ -945,7 +1018,7 @@ class ConnectionURIOpener(object):
         return conn
 
     def dbConnectionForScheme(self, scheme):
-        assert self.schemeBuilders.has_key(scheme), (
+        assert scheme in self.schemeBuilders, (
                "No SQLObject driver exists for %s (only %s)"
                % (scheme, ', '.join(self.schemeBuilders.keys())))
         return self.schemeBuilders[scheme]()
@@ -956,3 +1029,13 @@ registerConnection = TheURIOpener.registerConnection
 registerConnectionInstance = TheURIOpener.registerConnectionInstance
 connectionForURI = TheURIOpener.connectionForURI
 dbConnectionForScheme = TheURIOpener.dbConnectionForScheme
+
+# Register DB URI schemas
+import firebird
+import maxdb
+import mssql
+import mysql
+import postgres
+import rdbhost
+import sqlite
+import sybase
